#coding: utf-8

#https://github.com/ljatynu/NIMCGCN/tree/a0798ed29ae329dd71bff096ffc678527cc4099e/
from stanscofi.models import BasicModel
from stanscofi.preprocessing import CustomScaler

from benchscofi.implementations import NIMCGCNImplementation

import numpy as np
import random
import torch as t
from torch import optim
import os
from subprocess import Popen

def read_mat(df):
    return t.FloatTensor([np.asarray(df[irow,:],dtype=float).tolist() for irow in range(df.shape[0])])

class NIMCGCN(BasicModel):
    def __init__(self, params=None):
        params = params if (params is not None) else self.default_parameters()
        super(NIMCGCN, self).__init__(params)
        self.scalerS, self.scalerP = None, None
        self.validation=1 ## placeholder
        self.name = "NIMCGCN" 
        self.model = None

    def default_parameters(self):
        params = {
            "epoch" : 10,
            "alpha" : 0.2,
            "fg" : 256,
            "fd" : 256,
            "k" : 32,
            "display_epoch": 5,
            "random_state": 1234, 
            "learning_rate": 0.001,
        }
        return params

    def preprocessing(self, dataset, is_training=True, inf=2):
        if (self.scalerS is None):
            self.scalerS = CustomScaler(posinf=inf, neginf=-inf)
        S_ = self.scalerS.fit_transform(dataset.items.T.toarray().copy(), subset=None)
        S_ = np.nan_to_num(S_, nan=0.0) ##
        X_s = S_ if (S_.shape[0]==S_.shape[1]) else np.corrcoef(S_)
        if (self.scalerP is None):
            self.scalerP = CustomScaler(posinf=inf, neginf=-inf)
        P_ = self.scalerP.fit_transform(dataset.users.T.toarray().copy(), subset=None)
        P_ = np.nan_to_num(P_, nan=0.0) ##
        X_p = P_ if (P_.shape[0]==P_.shape[1]) else np.corrcoef(P_)
        A_sp = dataset.ratings.toarray().T # users x items
        ## https://github.com/ljatynu/NIMCGCN/tree/a0798ed29ae329dd71bff096ffc678527cc4099e
        ## do not take into account negative samples
        A_sp[A_sp==-1] = 0
        dataset_ = dict()
        dataset_['md_p'] = read_mat(A_sp)
        dataset_['md_true'] = read_mat(A_sp)

        zero_index = []
        one_index = []
        for i in range(dataset_['md_p'].size(0)):
            for j in range(dataset_['md_p'].size(1)):
                if dataset_['md_p'][i][j] < 1:
                    zero_index.append([i, j])
                if dataset_['md_p'][i][j] >= 1:
                    one_index.append([i, j])
        random.shuffle(one_index)
        random.shuffle(zero_index)
        zero_tensor = t.LongTensor(zero_index)
        one_tensor = t.LongTensor(one_index)
        dataset_['md'] = dict()
        dataset_['md']['train'] = [one_tensor, zero_tensor]

        dd_matrix = read_mat(X_s)
        dd_edge_index = NIMCGCNImplementation.get_edge_index(dd_matrix)
        dataset_['dd'] = {'data': dd_matrix, 'edge_index': dd_edge_index}

        mm_matrix = read_mat(X_p)
        mm_edge_index = NIMCGCNImplementation.get_edge_index(mm_matrix)
        dataset_['mm'] = {'data': mm_matrix, 'edge_index': mm_edge_index}

        sizes = NIMCGCNImplementation.Sizes(dataset_)
        dat = NIMCGCNImplementation.Dataset(self, dataset_)
        return [dat[0], X_s.shape[0], X_p.shape[0]] if (is_training) else [dat[0]]
        
    def model_fit(self, train_dat, d, m):
        t.manual_seed(self.random_state)
        np.random.seed(self.random_state) 
        random.seed(self.random_state)
        self.d, self.m = d, m
        self.model = NIMCGCNImplementation.Model(self)
        self.model.alpha = self.alpha
        self.model.epoch = self.epoch
        if (t.cuda.is_available()):
            model.cuda()
        optimizer = optim.Adam(self.model.parameters(), lr=self.learning_rate)
        NIMCGCNImplementation.train(self.model, train_dat, optimizer, self)

    def model_predict_proba(self, test_dat):
        outputs = self.model(test_dat)
        y_pred = outputs.detach().numpy().T
        return y_pred



#coding:utf-8

## source: https://raw.githubusercontent.com/jilljenn/vae/1d7f09af3bcaebfc5d8fa8cc18033d8bb8ca19bc/vfm-torch.py

import torch
from torch import nn, distributions
import numpy as np

#LINK = nn.functional.softplus
LINK = torch.abs

class CF(nn.Module):
    """
    Recommender system
    """
    def __init__(self, embedding_size, N_VARIATIONAL_SAMPLES, N=1, M=1, output='reg'): #N=nusers, M=nitems
        super().__init__()
        self.N, self.M = N, M
        self.embedding_size = embedding_size
        self.N_VARIATIONAL_SAMPLES = N_VARIATIONAL_SAMPLES
        self.output = output
        self.alpha = nn.Parameter(torch.Tensor([1e9]), requires_grad=True)
        self.global_bias_mean = nn.Parameter(torch.Tensor([0.]), requires_grad=True)
        self.global_bias_scale = nn.Parameter(torch.Tensor([1.]), requires_grad=True)
        self.prec_global_bias_prior = nn.Parameter(torch.Tensor([1.]), requires_grad=True)
        self.prec_user_bias_prior = nn.Parameter(torch.Tensor([1.]), requires_grad=True)
        self.prec_item_bias_prior = nn.Parameter(torch.Tensor([1.]), requires_grad=True)
        self.prec_user_entity_prior = nn.Parameter(torch.ones(self.embedding_size), requires_grad=True)
        self.prec_item_entity_prior = nn.Parameter(torch.ones(self.embedding_size), requires_grad=True)
        # self.alpha = torch.Tensor([1.])
        nn.init.uniform_(self.alpha)
        
        # bias_init = torch.cat((torch.randn(self.N + self.M, 1), torch.ones(self.N + self.M, 1) * (0.02 ** 0.5)), axis=1)
        # entity_init = torch.cat((
        #     torch.randn(self.N + self.M, self.embedding_size),
        #     torch.ones(self.N + self.M, self.embedding_size) * (0.02 ** 0.5),
        # ), axis=1)
        self.bias_params = nn.Embedding(self.N + self.M, 2)#.from_pretrained(bias_init)  # w
        self.entity_params = nn.Embedding(self.N + self.M, 2 * self.embedding_size)#.from_pretrained(entity_init)  # V

        self.saved_global_biases = []
        self.saved_mean_biases = []
        self.saved_mean_entities = []
        self.mean_saved_global_biases = None
        self.mean_saved_mean_biases = None
        self.mean_saved_mean_entities = None

        self.global_bias_prior = distributions.normal.Normal(0, 1)
        self.bias_prior = distributions.normal.Normal(0, 1)
        self.entity_prior = distributions.normal.Normal(0, 1)
        #     torch.zeros(self.N + self.M),
        #     torch.nn.functional.softplus(torch.cat((
        #         self.prec_user_bias_prior.repeat(N),
        #         self.prec_item_bias_prior.repeat(M)
        #     )))
        # )
        # self.entity_prior = distributions.normal.Normal(
        #     torch.zeros(self.N + self.M, self.embedding_size),
        #     torch.nn.functional.softplus(torch.cat((
        #         self.prec_user_entity_prior.repeat(N, 1),
        #         self.prec_item_entity_prior.repeat(M, 1)
        #     )))
        # )

    def save_weights(self):
        self.saved_global_biases.append(self.global_bias_mean.detach().numpy().copy())
        self.saved_mean_biases.append(self.bias_params.weight[:, 0].detach().numpy().copy())
        self.saved_mean_entities.append(self.entity_params.weight[:, :self.embedding_size].detach().numpy().copy())
        self.mean_saved_global_biases = np.array(self.saved_global_biases).mean(axis=0)
        self.mean_saved_mean_biases = np.array(self.saved_mean_biases).mean(axis=0)
        self.mean_saved_mean_entities = np.array(self.saved_mean_entities).mean(axis=0)
        # print('size of saved', np.array(self.saved_mean_entities).shape)
        # print('test', np.array(self.saved_mean_biases)[:3, 0])

    def forward(self, x):
        # print(("x",np.isnan(x.detach().numpy()).any()))
        uniq_entities, entity_pos, nb_occ_in_batch = torch.unique(x, return_inverse=True, return_counts=True)
        # print(("uniq_entities",np.isnan(uniq_entities.detach().numpy()).any()))
        uniq_users, nb_occ_user_in_batch = torch.unique(x[:, 0], return_counts=True)
        uniq_items, nb_occ_item_in_batch = torch.unique(x[:, 1], return_counts=True)
        # nb_uniq_users = len(uniq_users)
        # nb_uniq_items = len(uniq_items)
        # print('uniq', uniq_entities.shape, 'pos', entity_pos.shape)

        # self.global_bias_prior = distributions.normal.Normal(
        #     torch.Tensor([0.]), torch.nn.functional.softplus(self.prec_global_bias_prior))
        # Global bias
        global_bias_sampler = distributions.normal.Normal(
            self.global_bias_mean,
            LINK(self.global_bias_scale)
        )
        # Biases and entities
        bias_batch = self.bias_params(x)
        entity_batch = self.entity_params(x)
        uniq_bias_batch = self.bias_params(uniq_entities)#.reshape(-1, 2)
        uniq_entity_batch = self.entity_params(uniq_entities)#.reshape(-1, 2 * self.embedding_size)
        uniq_entity_batch = torch.nan_to_num(uniq_entity_batch, nan=1e-6) ###
        # print(("uniq_entity_batch",np.isnan(uniq_entity_batch.detach().numpy()).any()))
        # print('first', bias_batch.shape, entity_batch.shape)
        # print('samplers', uniq_bias_batch.shape, uniq_entity_batch.shape)
        # scale_bias = torch.ones_like(scale_bias) * 1e-6
        bias_sampler = distributions.normal.Normal(
            uniq_bias_batch[:, 0],
            LINK(uniq_bias_batch[:, 1])
        )
        # user_bias_posterior = distributions.normal.Normal(
        #     bias_batch[:, :, 0],
        #     LINK(bias_batch[:, :, 1])
        # )
        # diag_scale_entity = nn.functional.softplus(entity_batch[:, self.embedding_size:])
        # diag_scale_entity = torch.ones_like(diag_scale_entity) * 1e-6
        # print('scale entity', entity_batch.shape, scale_entity.shape)
        # print(("uniq_entity_batch2",np.isnan(uniq_entity_batch.detach().numpy()).any()))
        # print(("LINK uniq_entity_batch",np.isnan(LINK(uniq_entity_batch[:, self.embedding_size:]).detach().numpy()).any()))
        entity_sampler = distributions.normal.Normal(
            loc=uniq_entity_batch[:, :self.embedding_size],
            scale=LINK(uniq_entity_batch[:, self.embedding_size:])
        )
        # entity_posterior = distributions.normal.Normal(
        #     loc=entity_batch[:, :, :self.embedding_size],
        #     scale=LINK(entity_batch[:, :, self.embedding_size:])
        # )
        # self.entity_prior = distributions.normal.Normal(
        #     loc=torch.zeros_like(entity_batch[:, :, :self.embedding_size]),
        #     scale=torch.ones_like(entity_batch[:, :, :self.embedding_size])
        # )

        # print('batch shapes', entity_sampler.batch_shape, self.entity_prior.batch_shape)
        # print('event shapes', entity_sampler.event_shape, self.entity_prior.event_shape)
        global_bias = global_bias_sampler.rsample((self.N_VARIATIONAL_SAMPLES,))
        biases = bias_sampler.rsample((self.N_VARIATIONAL_SAMPLES,))#.reshape(
            # self.N_VARIATIONAL_SAMPLES, -1, 2)
        entities = entity_sampler.rsample((self.N_VARIATIONAL_SAMPLES,))#.reshape(
            # self.N_VARIATIONAL_SAMPLES, -1, 2, self.embedding_size)  # N_VAR_SAMPLES x BATCH_SIZE x 2 (user, item) x self.embedding_size
        # print('hola', biases.shape, entities.shape)
        sum_users_items_biases = biases[:, entity_pos].sum(axis=2).mean(axis=0).squeeze()
        users_items_emb = entities[:, entity_pos].prod(axis=2).sum(axis=2).mean(axis=0)
        # print('final', sum_users_items_biases.shape, users_items_emb.shape)

        if self.mean_saved_mean_biases is not None:
            last_global_bias = self.saved_global_biases[-1]
            last_bias_term = self.saved_mean_biases[-1][x].sum(axis=1).squeeze()
            last_embed_term = self.saved_mean_entities[-1][x].prod(axis=1).sum(axis=1)

            mean_global_bias = self.mean_saved_global_biases
            mean_bias_term = self.mean_saved_mean_biases[x].sum(axis=1).squeeze()
            mean_embed_term = self.mean_saved_mean_entities[x].prod(axis=1).sum(axis=1)
            # print(self.mean_saved_mean_biases[x].shape, mean_bias_term.shape)
            # print(self.mean_saved_mean_entities[x].shape, mean_embed_term.shape)
            last_logits = last_global_bias + last_bias_term + last_embed_term
            mean_logits = mean_global_bias + mean_bias_term + mean_embed_term #  + 
        else:
            last_logits = None
            mean_logits = None

        std_dev = torch.sqrt(1 / LINK(self.alpha))
        unscaled_pred = global_bias + sum_users_items_biases + users_items_emb

        if self.output == 'reg':
            likelihood = distributions.normal.Normal(unscaled_pred, std_dev)
        else:
            likelihood = distributions.bernoulli.Bernoulli(logits=unscaled_pred)
        # print('global bias sampler', global_bias_sampler)
        # print('global bias prior', self.global_bias_prior)
        # print('bias sampler', bias_sampler)
        # print('bias prior', self.bias_prior)
        # print('entity sampler', entity_sampler)
        # print('entity prior', self.entity_prior)
        # a = distributions.normal.Normal(torch.zeros(2, 3), torch.ones(2, 3))
        # b = distributions.normal.Normal(torch.zeros(2, 3), torch.ones(2, 3))
        # print('oh hey', distributions.kl.kl_divergence(a, b))
        # print('oh hey', distributions.kl.kl_divergence(entity_sampler, entity_sampler))
        # print('oh hiya', distributions.kl.kl_divergence(entity_sampler, self.entity_prior))
        # print('oh hey', distributions.kl.kl_divergence(self.entity_prior, self.entity_prior))

        # print(
        #     distributions.kl.kl_divergence(global_bias_sampler, self.global_bias_prior).shape,
        #     distributions.kl.kl_divergence(bias_sampler, self.bias_prior).sum(axis=1).shape,
        #     distributions.kl.kl_divergence(entity_sampler, self.entity_prior).sum(axis=[1, 2]).shape#.sum(axis=3).shape,#.sum(axis=2).shape
        # )

        kl_bias = distributions.kl.kl_divergence(bias_sampler, self.bias_prior)
        # print(('kl_bias', kl_bias.shape, kl_bias))
        # print('kl bias', kl_bias.shape)
        # print('bias sampler', bias_sampler)
        # print('entity sampler', entity_sampler)
        # print('entity prior', self.entity_prior)
        kl_entity = distributions.kl.kl_divergence(entity_sampler, self.entity_prior).sum(axis=1)
        # print('kl entity', kl_entity.shape)
        # print(('kl_entity', kl_entity.shape, kl_entity))

        nb_occ_in_train = nb_occ_in_batch[uniq_entities]
        nb_occ_user_in_train = nb_occ_in_train[uniq_users]
        nb_occ_item_in_train = nb_occ_in_train[uniq_items]
        # print(('nb_occ_in_train', nb_occ_in_train))
        # print(('nb_occ_in_batch', nb_occ_in_batch))
        # print(('nb_occ_in_train/nb_occ_in_batch', nb_occ_in_train/nb_occ_in_batch))
        # nb_occ_batch = torch.bincount(x.flatten())
        # print('nboccs', nb_occ_in_batch.shape, nb_occ_in_train.shape)
        # nb_occ_batch[x]

        user_normalizer = (nb_occ_user_in_batch / nb_occ_user_in_train).sum(axis=0)
        item_normalizer = (nb_occ_item_in_batch / nb_occ_item_in_train).sum(axis=0)
        # print('normalizers', user_normalizer.shape, item_normalizer.shape)

        # print('begin', ((kl_bias + kl_entity) * (nb_occ_in_batch / nb_occ_in_train)).shape)
        # print('ent', x)
        # print('ent', x <= self.N)
        # print('ent', (x <= self.N) * self.N)
        # print(('user_normalizer', user_normalizer.shape, user_normalizer.sum()))
        # print(('item_normalizer', item_normalizer.shape, item_normalizer.sum()))

        kl_rescaled = (
            (kl_bias + kl_entity) * (nb_occ_in_batch / nb_occ_in_train) *
            ((uniq_entities <= self.N) * self.N / user_normalizer + (uniq_entities > self.N) * self.M / item_normalizer)
        ).sum(axis=0)
        # print(('rescaled', kl_rescaled.shape, kl_rescaled))

        return (likelihood,
            last_logits, mean_logits,
            distributions.kl.kl_divergence(global_bias_sampler, self.global_bias_prior) +
            kl_rescaled
        )



#coding: utf-8

from benchscofi.SimpleBinaryClassifier import SimpleNeuralNetwork

import numpy as np

import tensorflow as tf
from tensorflow.python.keras.models import Model
from tensorflow.python.keras.layers import Input

class SimplePULearning(SimpleNeuralNetwork):
    def __init__(self, params=None):
        params = params if (params is not None) else self.default_parameters()
        super(SimplePULearning, self).__init__(params)
        self.name = "SimplePULearning"

    def default_parameters(self):
        params = super(SimplePULearning, self).default_parameters()
        params.setdefault("PI", 0.33)
        return params

    def _custom_cross_entropy(self, log_ratio_p, labels):
        temp = tf.cast(tf.math.softplus(log_ratio_p), np.float32)
        weights = tf.cast(tf.constant([-1 / (self.PI - 1), (2 * self.PI - 1) / (self.PI - 1)]), np.float32)
        coef = tf.constant([0., 1.])
        bundle = temp[..., None] * weights[None, ...] - log_ratio_p[..., None] * coef[None, ...]
        oh = tf.one_hot(labels, depth=2)
        return tf.reduce_sum(tf.reduce_sum(bundle * oh, axis=0) / tf.reduce_sum(oh, axis=0))

    def nn_creation(self, x_p, x_q, log_ratio_p, log_ratio_q, X, y):
        labels = Input(shape=(1,), dtype=tf.int32)
        model = Model(inputs=[x_p, labels], outputs=log_ratio_p)
        model.add_loss(self._custom_cross_entropy(log_ratio_p, labels))
        model.compile(optimizer='rmsprop', loss=None, metrics=['accuracy'])
        XX = np.concatenate(tuple([X[y==v,:] for v in [1,-1]]), axis=0)
        YY = np.concatenate(
            (
                tf.ones(np.sum(y==1)), tf.zeros(np.sum(y==-1))
            )        
        ).astype(np.int32)
        return model, [XX, YY], YY

    def model_predict_proba(self, X):
        preds = self.model.predict(x=[X,tf.zeros(X.shape[0])])
        return np.ravel(tf.nn.sigmoid(preds).numpy())



from setuptools import setup, find_packages

NAME = "benchscofi"
VERSION = "1.0.0"

setup(name=NAME,
    version=VERSION,
    author="Clémence Réda",
    author_email="recess-project@proton.me",
    url="https://github.com/RECeSS-EU-Project/benchscofi",
    license_files = ('LICENSE'),
    classifiers=[
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "License :: OSI Approved :: MIT License",
    ],
    keywords='',
    description="Package which contains implementations of published collaborative filtering-based algorithms for drug repurposing.",
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    packages=find_packages(where="src"),
    package_dir={'':"src"},
    python_requires='>=3.8.5',
    install_requires=[
        "stanscofi>=2.0.0",
        "tensorflow==2.4.3",
        "pulearn>=0.0.7",
        "torch>=2.0.1",
        "fastai>=2.7.12",
        "torch_geometric>=2.3.1",
        "pyFFM",
        "protobuf==3.9.*",
        "pydantic==1.9.*",
        "pytorch-lightning<2.0.0",
        "torchmetrics",
        "numpy==1.19.4",
        "scikit-learn==1.2.*",
        "dgl>=1.1.1",
        "libmf>=0.9.2",
    ],
    entry_points={},
)



#coding:utf-8

## source: https://ethen8181.github.io/machine-learning/recsys/4_bpr.html

import sys
import numpy as np
from tqdm import trange
from sklearn.preprocessing import normalize
from sklearn.neighbors import NearestNeighbors

bpr_params = {
    'reg': 0.01,
    'learning_rate': 0.1,
    'n_iters': 160,
    'n_factors': 15,
    'batch_size': 100
}

class BPR:
    """
    Bayesian Personalized Ranking (BPR) for implicit feedback data

    Parameters
    ----------
    learning_rate : float, default 0.01
        learning rate for gradient descent

    n_factors : int, default 20
        Number/dimension of user and item latent factors

    n_iters : int, default 15
        Number of iterations to train the algorithm
        
    batch_size : int, default 1000
        batch size for batch gradient descent, the original paper
        uses stochastic gradient descent (i.e., batch size of 1),
        but this can make the training unstable (very sensitive to
        learning rate)

    reg : int, default 0.01
        Regularization term for the user and item latent factors

    seed : int, default 1234
        Seed for the randomly initialized user, item latent factors

    verbose : bool, default True
        Whether to print progress bar while training

    Attributes
    ----------
    user_factors : 2d ndarray, shape [n_users, n_factors]
        User latent factors learnt

    item_factors : 2d ndarray, shape [n_items, n_factors]
        Item latent factors learnt

    References
    ----------
    S. Rendle, C. Freudenthaler, Z. Gantner, L. Schmidt-Thieme 
    Bayesian Personalized Ranking from Implicit Feedback
    - https://arxiv.org/abs/1205.2618
    """
    def __init__(self, learning_rate = 0.01, n_factors = 15, n_iters = 10, 
                 batch_size = 1000, reg = 0.01, seed = 1234, verbose = True):
        self.reg = reg
        self.seed = seed
        self.verbose = verbose
        self.n_iters = n_iters
        self.n_factors = n_factors
        self.batch_size = batch_size
        self.learning_rate = learning_rate
        
        # to avoid re-computation at predict
        self._prediction = None
        
    def fit(self, ratings):
        """
        Parameters
        ----------
        ratings : scipy sparse csr_matrix, shape [n_users, n_items]
            sparse matrix of user-item interactions
        """
        indptr = ratings.indptr
        indices = ratings.indices
        n_users, n_items = ratings.shape
        
        # ensure batch size makes sense, since the algorithm involves
        # for each step randomly sample a user, thus the batch size
        # should be smaller than the total number of users or else
        # we would be sampling the user with replacement
        batch_size = self.batch_size
        if n_users < batch_size:
            batch_size = n_users
            sys.stderr.write('WARNING: Batch size is greater than number of users,'
                             'switching to a batch size of {}\n'.format(n_users))

        batch_iters = n_users // batch_size
        
        # initialize random weights
        rstate = np.random.RandomState(self.seed)
        self.user_factors = rstate.normal(size = (n_users, self.n_factors))
        self.item_factors = rstate.normal(size = (n_items, self.n_factors))
        
        # progress bar for training iteration if verbose is turned on
        loop = range(self.n_iters)
        if self.verbose:
            loop = trange(self.n_iters, desc = self.__class__.__name__)
        
        for _ in loop:
            for _ in range(batch_iters):
                sampled = self._sample(n_users, n_items, indices, indptr)
                sampled_users, sampled_pos_items, sampled_neg_items = sampled
                self._update(sampled_users, sampled_pos_items, sampled_neg_items)

        return self
    
    def _sample(self, n_users, n_items, indices, indptr):
        """sample batches of random triplets u, i, j"""
        sampled_pos_items = np.zeros(self.batch_size, dtype = np.int)
        sampled_neg_items = np.zeros(self.batch_size, dtype = np.int)
        sampled_users = np.random.choice(
            n_users, size = self.batch_size, replace = False)

        for idx, user in enumerate(sampled_users):
            pos_items = indices[indptr[user]:indptr[user + 1]]
            if (len(pos_items)==0): ##
                continue ##
            pos_item = np.random.choice(pos_items)
            neg_item = np.random.choice(n_items)
            while neg_item in pos_items:
                neg_item = np.random.choice(n_items)

            sampled_pos_items[idx] = pos_item
            sampled_neg_items[idx] = neg_item

        return sampled_users, sampled_pos_items, sampled_neg_items
                
    def _update(self, u, i, j):
        """
        update according to the bootstrapped user u, 
        positive item i and negative item j
        """
        user_u = self.user_factors[u]
        item_i = self.item_factors[i]
        item_j = self.item_factors[j]
        
        # decompose the estimator, compute the difference between
        # the score of the positive items and negative items; a
        # naive implementation might look like the following:
        # r_ui = np.diag(user_u.dot(item_i.T))
        # r_uj = np.diag(user_u.dot(item_j.T))
        # r_uij = r_ui - r_uj
        
        # however, we can do better, so
        # for batch dot product, instead of doing the dot product
        # then only extract the diagonal element (which is the value
        # of that current batch), we perform a hadamard product, 
        # i.e. matrix element-wise product then do a sum along the column will
        # be more efficient since it's less operations
        # http://people.revoledu.com/kardi/tutorial/LinearAlgebra/HadamardProduct.html
        # r_ui = np.sum(user_u * item_i, axis = 1)
        #
        # then we can achieve another speedup by doing the difference
        # on the positive and negative item up front instead of computing
        # r_ui and r_uj separately, these two idea will speed up the operations
        # from 1:14 down to 0.36
        r_uij = np.sum(user_u * (item_i - item_j), axis = 1)
        sigmoid = np.exp(-r_uij) / (1.0 + np.exp(-r_uij))
        
        # repeat the 1 dimension sigmoid n_factors times so
        # the dimension will match when doing the update
        sigmoid_tiled = np.tile(sigmoid, (self.n_factors, 1)).T

        # update using gradient descent
        grad_u = sigmoid_tiled * (item_j - item_i) + self.reg * user_u
        grad_i = sigmoid_tiled * -user_u + self.reg * item_i
        grad_j = sigmoid_tiled * user_u + self.reg * item_j
        self.user_factors[u] -= self.learning_rate * grad_u
        self.item_factors[i] -= self.learning_rate * grad_i
        self.item_factors[j] -= self.learning_rate * grad_j
        return self

    def predict(self):
        """
        Obtain the predicted ratings for every users and items
        by doing a dot product of the learnt user and item vectors.
        The result will be cached to avoid re-computing it every time
        we call predict, thus there will only be an overhead the first
        time we call it. Note, ideally you probably don't need to compute
        this as it returns a dense matrix and may take up huge amounts of
        memory for large datasets
        """
        if self._prediction is None:
            self._prediction = self.user_factors.dot(self.item_factors.T)

        return self._prediction

    def _predict_user(self, user):
        """
        returns the predicted ratings for the specified user,
        this is mainly used in computing evaluation metric
        """
        user_pred = self.user_factors[user].dot(self.item_factors.T)
        return user_pred

    def recommend(self, ratings, N = 5):
        """
        Returns the top N ranked items for given user id,
        excluding the ones that the user already liked
        
        Parameters
        ----------
        ratings : scipy sparse csr_matrix, shape [n_users, n_items]
            sparse matrix of user-item interactions 
        
        N : int, default 5
            top-N similar items' N
        
        Returns
        -------
        recommendation : 2d ndarray, shape [number of users, N]
            each row is the top-N ranked item for each query user
        """
        n_users = ratings.shape[0]
        recommendation = np.zeros((n_users, N), dtype = np.uint32)
        for user in range(n_users):
            top_n = self._recommend_user(ratings, user, N)
            recommendation[user] = top_n

        return recommendation

    def _recommend_user(self, ratings, user, N):
        """the top-N ranked items for a given user"""
        scores = self._predict_user(user)

        # compute the top N items, removing the items that the user already liked
        # from the result and ensure that we don't get out of bounds error when 
        # we ask for more recommendations than that are available
        liked = set(ratings[user].indices)
        count = N + len(liked)
        if count < scores.shape[0]:

            # when trying to obtain the top-N indices from the score,
            # using argpartition to retrieve the top-N indices in 
            # unsorted order and then sort them will be faster than doing
            # straight up argort on the entire score
            # http://stackoverflow.com/questions/42184499/cannot-understand-numpy-argpartition-output
            ids = np.argpartition(scores, -count)[-count:]
            best_ids = np.argsort(scores[ids])[::-1]
            best = ids[best_ids]
        else:
            best = np.argsort(scores)[::-1]

        top_n = list(islice((rec for rec in best if rec not in liked), N))
        return top_n
    
    def get_similar_items(self, N = 5, item_ids = None):
        """
        return the top N similar items for itemid, where
        cosine distance is used as the distance metric
        
        Parameters
        ----------
        N : int, default 5
            top-N similar items' N
            
        item_ids : 1d iterator, e.g. list or numpy array, default None
            the item ids that we wish to find the similar items
            of, the default None will compute the similar items
            for all the items
        
        Returns
        -------
        similar_items : 2d ndarray, shape [number of query item_ids, N]
            each row is the top-N most similar item id for each
            query item id
        """
        # cosine distance is proportional to normalized euclidean distance,
        # thus we normalize the item vectors and use euclidean metric so
        # we can use the more efficient kd-tree for nearest neighbor search;
        # also the item will always to nearest to itself, so we add 1 to 
        # get an additional nearest item and remove itself at the end
        normed_factors = normalize(self.item_factors)
        knn = NearestNeighbors(n_neighbors = N + 1, metric = 'euclidean')
        knn.fit(normed_factors)

        # returns a distance, index tuple,
        # we don't actually need the distance
        if item_ids is not None:
            normed_factors = normed_factors[item_ids]

        _, items = knn.kneighbors(normed_factors)
        similar_items = items[:, 1:].astype(np.uint32)
        return similar_items


#coding:utf-8

## https://raw.githubusercontent.com/MrChrisJohnson/logistic-mf/90d4d7d44fcd0f373fbb5a827c2dc9f5da68d84b/logistic_mf.py

import time
import numpy as np


def load_matrix(filename, num_users, num_items):
    t0 = time.time()
    counts = np.zeros((num_users, num_items))
    total = 0.0
    num_zeros = num_users * num_items
    for i, line in enumerate(open(filename, 'r')):
        user, item, count = line.strip().split('\t')
        user = int(user)
        item = int(item)
        count = float(count)
        counts[user][item] = count
        total += count
        num_zeros -= 1
    alpha = num_zeros / total
    print('alpha %.2f' % alpha)
    counts *= alpha
    t1 = time.time()
    print('Finished loading matrix in %f seconds' % (t1 - t0))
    return counts


class LogisticMF():

    def __init__(self, counts, num_factors, reg_param=0.6, gamma=1.0,
                 iterations=30):
        self.counts = counts
        self.num_users = counts.shape[0]
        self.num_items = counts.shape[1]
        self.num_factors = num_factors
        self.iterations = iterations
        self.reg_param = reg_param
        self.gamma = gamma

    def train_model(self):

        self.ones = np.ones((self.num_users, self.num_items))
        self.user_vectors = np.random.normal(size=(self.num_users,
                                                   self.num_factors))
        self.item_vectors = np.random.normal(size=(self.num_items,
                                                   self.num_factors))
        self.user_biases = np.random.normal(size=(self.num_users, 1))
        self.item_biases = np.random.normal(size=(self.num_items, 1))

        user_vec_deriv_sum = np.zeros((self.num_users, self.num_factors))
        item_vec_deriv_sum = np.zeros((self.num_items, self.num_factors))
        user_bias_deriv_sum = np.zeros((self.num_users, 1))
        item_bias_deriv_sum = np.zeros((self.num_items, 1))
        for i in range(self.iterations):
            t0 = time.time()
            # Fix items and solve for users
            # take step towards gradient of deriv of log likelihood
            # we take a step in positive direction because we are maximizing LL
            user_vec_deriv, user_bias_deriv = self.deriv(True)
            user_vec_deriv_sum += np.square(user_vec_deriv)
            user_bias_deriv_sum += np.square(user_bias_deriv)
            vec_step_size = self.gamma / np.sqrt(user_vec_deriv_sum)
            bias_step_size = self.gamma / np.sqrt(user_bias_deriv_sum)
            self.user_vectors += vec_step_size * user_vec_deriv
            self.user_biases += bias_step_size * user_bias_deriv

            # Fix users and solve for items
            # take step towards gradient of deriv of log likelihood
            # we take a step in positive direction because we are maximizing LL
            item_vec_deriv, item_bias_deriv = self.deriv(False)
            item_vec_deriv_sum += np.square(item_vec_deriv)
            item_bias_deriv_sum += np.square(item_bias_deriv)
            vec_step_size = self.gamma / np.sqrt(item_vec_deriv_sum)
            bias_step_size = self.gamma / np.sqrt(item_bias_deriv_sum)
            self.item_vectors += vec_step_size * item_vec_deriv
            self.item_biases += bias_step_size * item_bias_deriv
            t1 = time.time()

            print('iteration %i finished in %f seconds' % (i + 1, t1 - t0))

    def deriv(self, user):
        if user:
            vec_deriv = np.dot(self.counts, self.item_vectors)
            bias_deriv = np.expand_dims(np.sum(self.counts, axis=1), 1)

        else:
            vec_deriv = np.dot(self.counts.T, self.user_vectors)
            bias_deriv = np.expand_dims(np.sum(self.counts, axis=0), 1)
        A = np.dot(self.user_vectors, self.item_vectors.T)
        A += self.user_biases
        A += self.item_biases.T
        A = np.exp(A)
        A /= (A + self.ones)
        A = (self.counts + self.ones) * A

        if user:
            vec_deriv -= np.dot(A, self.item_vectors)
            bias_deriv -= np.expand_dims(np.sum(A, axis=1), 1)
            # L2 regularization
            vec_deriv -= self.reg_param * self.user_vectors
        else:
            vec_deriv -= np.dot(A.T, self.user_vectors)
            bias_deriv -= np.expand_dims(np.sum(A, axis=0), 1)
            # L2 regularization
            vec_deriv -= self.reg_param * self.item_vectors
        return (vec_deriv, bias_deriv)

    def log_likelihood(self):
        loglik = 0
        A = np.dot(self.user_vectors, self.item_vectors.T)
        A += self.user_biases
        A += self.item_biases.T
        B = A * self.counts
        loglik += np.sum(B)

        A = np.exp(A)
        A += self.ones

        A = np.log(A)
        A = (self.counts + self.ones) * A
        loglik -= np.sum(A)

        # L2 regularization
        loglik -= 0.5 * self.reg_param * np.sum(np.square(self.user_vectors))
        loglik -= 0.5 * self.reg_param * np.sum(np.square(self.item_vectors))
        return loglik

    def print_vectors(self):
        user_vecs_file = open('logmf-user-vecs-%i' % self.num_factors, 'w')
        for i in range(self.num_users):
            vec = ' '.join(map(str, self.user_vectors[i]))
            line = '%i\t%s\n' % (i, vec)
            user_vecs_file.write(line)
        user_vecs_file.close()
        item_vecs_file = open('logmf-item-vecs-%i' % self.num_factors, 'w')
        for i in range(self.num_items):
            vec = ' '.join(map(str, self.item_vectors[i]))
            line = '%i\t%s\n' % (i, vec)
            item_vecs_file.write(line)
        item_vecs_file.close()



#coding: utf-8

## Minimal example of model in stanscofi

from stanscofi.models import BasicModel
import numpy as np

class Constant(BasicModel):
    '''
    A example model which always predicts the positive class. It is a subclass of stanscofi.models.BasicModel (please refer to the documentation of this class for more information)

    ...

    Parameters
    ----------
    params : dict
        empty

    Attributes
    ----------
    name : str
        the name of the model
    ...
        other attributes might be present depending on the type of model

    Methods
    -------
    Same as BasicModel class
    __init__(params)
        Initialize the model with preselected parameters
    default_parameters()
        Outputs a dictionary which contains default values of parameters
    model_fit()
        Preprocess and fit the model (not implemented in BasicModel)
    model_predict_proba()
        Outputs predictions of the fitted model on test_dataset (not implemented in BasicModel)
    '''
    def __init__(self, params=None):
        '''
        Creates an instance of benchscofi.Constant

        ...

        Parameters
        ----------
        params : dict
            dictionary which contains a key called "decision_threshold" with a float value which determines the decision threshold to label a positive class
        '''
        params = params if (params is not None) else self.default_parameters()
        super(Constant, self).__init__(params)
        self.name = "Constant"

    def default_parameters(self):
        params = {}
        return params

    def preprocessing(self, dataset, is_training=True):
        '''
        Preprocessing step, which is empty for this model.
        The general rule is that, for training:
         - if the algorithm takes as input a matrix, then consider all ratings (y in {-1,0,1})
         - if the algorithm takes as input a collection pairs (item, user), then consider only known ratings (y in {-1,1})
        For testing/predicting, regardless of the type of input considered by the algorithm, consider all ratings (y in {-1,0,1})

        ...

        Parameters
        ----------
        dataset : stanscofi.Dataset
            dataset to convert

        Returns
        ----------
        res : list of a single stanscofi.Dataset
            input to the fitting method (should be a list of all arguments)
        '''
        return [dataset]
        
    def model_fit(self, train_dataset):
        '''
        Fitting the Constant model on the training dataset (no fitting step here).

        ...

        Parameters
        ----------
        train_dataset : stanscofi.Dataset
            training dataset on which the model should fit
        '''
        pass

    def model_predict_proba(self, test_dataset):
        '''
        Making predictions using the Constant model on the testing dataset.

        ...

        Parameters
        ----------
        test_dataset : stanscofi.Dataset
            testing dataset on which the model should be validated
        '''
        scores = np.ones(test_dataset.ratings.shape)
        return scores 



#coding: utf-8

from stanscofi.models import BasicModel
from stanscofi.preprocessing import preprocessing_XY

from pyffm import PyFFM ## FM not tested!
import pandas as pd
import numpy as np

class FFMWrapper(BasicModel):
    def __init__(self, params=None):
        params = params if (params is not None) else self.default_parameters()
        super(FFMWrapper, self).__init__(params)
        assert self.preprocessing_str in ["Perlman_procedure", "meanimputation_standardize", "same_feature_preprocessing"]
        self.model = PyFFM(model='ffm', 
            training_params={"epochs": self.epochs, "reg_lambda": self.reg_lambda}
        )
        self.scalerS, self.scalerP, self.filter = None, None, None
        self.name = "FFMWrapper"

    def default_parameters(self):
        params = {
            "epochs": 2, #5,
            "reg_lambda": 0.002,
            "preprocessing_str": "meanimputation_standardize", "subset": None,
        }
        return params

    def preprocessing(self, dataset, is_training=True):
        X, y, scalerS, scalerP, filter_ = preprocessing_XY(dataset, self.preprocessing_str, subset_=self.subset, filter_=self.filter, scalerS=self.scalerS, scalerP=self.scalerP, inf=2, njobs=1)
        self.filter = filter_
        self.scalerS = scalerS
        self.scalerP = scalerP
        keep_ids = (y!=0)
        df = pd.DataFrame(np.concatenate((np.asarray(y[keep_ids]>0, dtype=int).reshape((np.sum(keep_ids),1)), X[keep_ids,:]), axis=1), index=range(np.sum(keep_ids)), columns=["click"]+list(map(str,range(X.shape[1]))))
        return [df] if (is_training) else [df, keep_ids]

    def model_fit(self, df):
        self.model.train(df, label_name="click")

    def model_predict_proba(self, df, keep_ids):
        preds = self.model.predict(df.drop(columns=["click"]))
        scores = np.zeros(keep_ids.shape)
        scores[keep_ids] = preds
        return scores



# coding:utf-8

alswr_params = {
    "n_iters":15, "n_factors":15, "alpha":15., "reg":0.01,
}

import numpy as np
from tqdm import trange

##https://ethen8181.github.io/machine-learning/recsys/2_implicit.html#Implementation
class ALSWR:
    """
    Alternating Least Squares with Weighted Regularization
    for implicit feedback

    Parameters
    ----------
    n_iters : int
        number of iterations to train the algorithm

    n_factors : int
        number/dimension of user and item latent factors

    alpha : int
        scaling factor that indicates the level of confidence in preference

    reg : int
        regularization term for the user and item latent factors

    seed : int
        seed for the randomly initialized user, item latent factors

    Reference
    ---------
    Y. Hu, Y. Koren, C. Volinsky Collaborative Filtering for Implicit Feedback Datasets
    http://yifanhu.net/PUB/cf.pdf
    """
    def __init__(self, n_iters, n_factors, alpha, reg, seed):
        self.reg = reg
        self.seed = seed
        self.alpha = alpha
        self.n_iters = n_iters
        self.n_factors = n_factors
    
    def fit(self, ratings):
        """
        ratings : scipy sparse csr_matrix [n_users, n_items]
            sparse matrix of user-item interactions
        """        
        # the original confidence vectors should include a + 1,
        # but this direct addition is not allowed when using sparse matrix,
        # thus we'll have to deal with this later in the computation
        Cui = ratings.copy().tocsr()
        Cui.data *= self.alpha
        Ciu = Cui.T.tocsr()
        self.n_users, self.n_items = Cui.shape
        
        # initialize user latent factors and item latent factors
        # randomly with a specified set seed
        rstate = np.random.RandomState(self.seed)
        self.user_factors = rstate.normal(size = (self.n_users, self.n_factors))
        self.item_factors = rstate.normal(size = (self.n_items, self.n_factors))
        
        for _ in trange(self.n_iters, desc = 'training progress'):
            self._als_step(Cui, self.user_factors, self.item_factors)
            self._als_step(Ciu, self.item_factors, self.user_factors)  
        
        return self
    
    def _als_step(self, Cui, X, Y):
        """
        when solving the user latent vectors,
        the item vectors will be fixed and vice versa
        """
        # the variable name follows the notation when holding
        # the item vector Y constant and solving for user vector X
        
        # YtY is a d * d matrix that is computed
        # independently of each user
        YtY = Y.T.dot(Y)
        data = Cui.data
        indptr, indices = Cui.indptr, Cui.indices

        # for every user build up A and b then solve for Ax = b,
        # this for loop is the bottleneck and can be easily parallized
        # as each users' computation is independent of one another
        for u in range(self.n_users):
            # initialize a new A and b for every user
            b = np.zeros(self.n_factors)
            A = YtY + self.reg * np.eye(self.n_factors)
            
            for index in range(indptr[u], indptr[u + 1]):
                # indices[index] stores non-zero positions for a given row
                # data[index] stores corresponding confidence,
                # we also add 1 to the confidence, since we did not 
                # do it in the beginning, when we were to give every 
                # user-item pair and minimal confidence
                i = indices[index]
                confidence = data[index] + 1
                factor = Y[i]

                # for b, Y^T C^u p_u
                # there should be a times 1 for the preference 
                # Pui = 1
                # b += confidence * Y[i] * Pui
                # but the times 1 can be dropped
                b += confidence * factor
                
                # for A, Y^T (C^u - I) Y
                A += (confidence - 1) * np.outer(factor, factor)

            X[u] = np.linalg.solve(A, b)
        
        return self

    def predict(self):
        """predict ratings for every user and item"""
        prediction = self.user_factors.dot(self.item_factors.T)
        return prediction
    
    def _predict_user(self, user):
        """
        returns the predicted ratings for the specified user,
        this is mainly used in computing evaluation metric
        """
        user_pred = self.user_factors[user].dot(self.item_factors.T)
        return user_pred
